import ast
from typing import List, Tuple, TypeVar, Optional

from asttokens import ASTTokens  # type: ignore[import]

from .ast import Position, NodeFinder, get_current_indent
from .editing import (
    coalesce,
    Insertion,
    INDENT_SIZE,
    MutationType,
    WrappingSummary,
    merge_insertions,
    indent_interim_lines,
)
from .wrappers import WRAPPING_FUNTIONS
from .exceptions import TuckError

TAst = TypeVar('TAst', bound=ast.AST)

WRAPPABLE_NODE_TYPES = tuple(x for x, _ in WRAPPING_FUNTIONS)


class TargetSyntaxError(TuckError):
    def __init__(self, inner: SyntaxError) -> None:
        super().__init__('target_syntax_error', str(inner))


def get_wrapping_summary(asttokens: ASTTokens, node: ast.AST) -> WrappingSummary:
    for ast_type, func in WRAPPING_FUNTIONS:
        if isinstance(node, ast_type):
            return func(asttokens, node)

    if not isinstance(node, WRAPPABLE_NODE_TYPES):
        raise AssertionError("Unable to get wrapping positions for {}".format(node))

    raise AssertionError("Unsupported node type {}".format(node))


def remove_redunant_wrapping_operations(
    asttokens: ASTTokens,
    wrapping_summary: WrappingSummary,
) -> WrappingSummary:
    """
    The initial wrapping summary defines where a correctly tucked statement
    should have various things, however we also need to cope with some or all of
    them already being present.

    This util inspects the actual token stream and removes anything from our
    summary which doesn't need to happen.
    """

    def should_keep(
        position: Position,
        mutation: MutationType,
        previous: Optional[Tuple[Position, MutationType]],
    ) -> bool:
        if mutation == MutationType.TRAILING_COMMA:
            tok = asttokens.get_token(position.line, position.col)
            prev_token = asttokens.prev_token(tok)

            if prev_token.string == ',':
                return False

        elif mutation in (MutationType.WRAP, MutationType.WRAP_INDENT):
            tok = asttokens.get_token(position.line, position.col)
            prev_token = asttokens.prev_token(tok, include_extra=True)

            if prev_token.string == '\n':
                return False

        elif mutation == MutationType.INDENT and previous is not None:
            prev_position, prev_mutation = previous

            if prev_mutation in (
                MutationType.WRAP,
                MutationType.WRAP_INDENT,
            ):
                if prev_position == position:
                    return False

        return True

    wrapping_summary = [
        current
        for prev, current in zip([None, *wrapping_summary], wrapping_summary)
        if should_keep(*current, prev)
    ]

    return wrapping_summary


def determine_insertions(asttokens: ASTTokens, position: Position) -> List[Insertion]:
    finder = NodeFinder(position, WRAPPABLE_NODE_TYPES)
    finder.visit(asttokens.tree)

    node = finder.found_node

    # Note: insertions are actually applied in reverse, though we'll generate
    # them forwards.

    current_indent = get_current_indent(asttokens, node)

    mutations = {
        MutationType.WRAP: "\n" + " " * current_indent,
        MutationType.WRAP_INDENT: "\n" + " " * (current_indent + INDENT_SIZE),
        MutationType.INDENT: " " * INDENT_SIZE,
        MutationType.TRAILING_COMMA: ",",
        MutationType.OPEN_PAREN: "(",
        MutationType.CLOSE_PAREN: ")",
    }

    wrapping_summary = get_wrapping_summary(asttokens, node)

    wrapping_summary = indent_interim_lines(asttokens, wrapping_summary)

    wrapping_summary = remove_redunant_wrapping_operations(asttokens, wrapping_summary)

    insertions = [
        (pos, ''.join(mutations[x] for x in mutation_types))
        for pos, mutation_types in coalesce(wrapping_summary)
    ]

    return insertions


def process(
    positions: List[Position],
    content: str,
    filename: str,
) -> List[Insertion]:
    try:
        asttokens = ASTTokens(content, parse=True, filename=filename)
    except SyntaxError as e:
        # Catch syntax errors within the file we were asked to parse. We trust
        # that the error is not within asttokens itself.
        raise TargetSyntaxError(e) from e

    insertions = merge_insertions(
        determine_insertions(asttokens, position)
        for position in positions
    )

    return insertions



import sys
import json
import difflib
import argparse
from typing import Dict, List, Union

from .ast import Position
from .main import process
from .editing import Insertion, apply_insertions
from .exceptions import TuckError

FAIL = '\033[91m'
ENDC = '\033[0m'


LSP_Range = Dict[str, Dict[str, int]]
LSP_TextEdit = Dict[str, Union[str, LSP_Range]]


def insertion_as_lsp_data(position: Position, new_text: str) -> LSP_TextEdit:
    """
    Convert an expanded `Insertion` to a Language Server Protocol compatible
    dictionaries for display as JSON.

    Note: in LSP line numbers are zero-based, while our `Position`s are
    one-based.
    """
    pos = {'line': position.line - 1, 'character': position.col}
    return {
        'range': {'start': pos, 'end': pos},
        'newText': new_text,
    }


def print_edits(insertions: List[Insertion]) -> None:
    edits = [insertion_as_lsp_data(*x) for x in insertions]
    print(json.dumps({'edits': edits}))


def parse_position(position: str) -> Position:
    line, col = [int(x) for x in position.split(':')]
    return Position(line, col)


def parse_args(argv: List[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Wrap the python statement at a given position within a text document.",
    )
    parser.add_argument(
        'file',
        type=argparse.FileType(mode='r'),
        help="The file to read from. Use '-' to read from STDIN.",
    )
    parser.add_argument(
        '--positions',
        required=True,
        nargs='+',
        type=parse_position,
        help=(
            "The positions within the file to wrap at. "
            "Express in LINE:COL format, with 1-based line numbers. "
            "When multiple locations are specified, they must appear within "
            "distinct statements. It is an error if the edits overlap."
        ),
    )
    parser.add_argument('--mode', choices=('wrap', 'unwrap'), default='wrap')
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--diff',
        action='store_true',
        help="Print the changes as a unified diff rather than printing the new content.",
    )
    group.add_argument(
        '--edits',
        action='store_true',
        help=(
            "Print the changes as language-server-protocol compatible edits, "
            "rather than printing the new content."
        ),
    )
    return parser.parse_args(argv)


def run(args: argparse.Namespace) -> None:
    with args.file:
        content = args.file.read()

    try:
        insertions = process(args.positions, content, args.file.name)
    except TuckError as e:
        if args.edits:
            # Assume the consumer is a tool looking for JSON
            print(
                json.dumps({'error': {'code': e.code, 'message': e.message}}),
                file=sys.stderr,
            )
        else:
            print(FAIL + e.message + ENDC, file=sys.stderr)
        return

    if args.diff:
        new_content = apply_insertions(content, insertions)
        print(
            "".join(difflib.unified_diff(
                content.splitlines(keepends=True),
                new_content.splitlines(keepends=True),
                'original',
                'formatted',
            )),
            end='',
        )
    elif args.edits:
        print_edits(insertions)
    else:
        new_content = apply_insertions(content, insertions)
        print(new_content, end='')


def main(argv: List[str] = sys.argv[1:]) -> None:
    return run(parse_args(argv))



import ast
import token
import functools
from typing import List, Type, Tuple, TypeVar, Optional

import asttokens.util  # type: ignore[import]
from asttokens import ASTTokens

from .exceptions import TuckError

TAst = TypeVar('TAst', bound=ast.AST)


def _first_token(node: ast.AST) -> asttokens.util.Token:
    if isinstance(node, ast.GeneratorExp):
        return _first_token(node.elt)

    return node.first_token  # type: ignore[attr-defined]


def _last_token(node: ast.AST) -> asttokens.util.Token:
    if isinstance(node, ast.GeneratorExp):
        return _last_token(node.generators[-1])

    return node.last_token  # type: ignore[attr-defined]


@functools.total_ordering
class Position:
    """
    A position within a document, compatible with Python AST positions.

    Line numbers are one-based, columns are zero-based.
    """

    @classmethod
    def from_node_start(cls, node: ast.AST) -> 'Position':
        return cls(*_first_token(node).start)

    @classmethod
    def from_node_end(cls, node: ast.AST) -> 'Position':
        return cls(*_last_token(node).start)

    def __init__(self, line: int, col: int) -> None:
        self.line = line
        self.col = col

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Position):
            return NotImplemented

        return self.line == other.line and self.col == other.col

    def __lt__(self, other: 'Position') -> bool:
        if not isinstance(other, Position):
            return NotImplemented  # type: ignore  # unreachable

        if self.line < other.line:
            return True

        if self.line > other.line:
            return False

        return self.col < other.col

    def __repr__(self) -> str:
        return 'Position(line={}, col={})'.format(self.line, self.col)


class NodeSearchError(TuckError, ValueError):
    """Base type for all node search errors."""


class NoNodeFoundError(NodeSearchError):
    def __init__(self) -> None:
        super().__init__('no_node_found', "No AST nodes were found!")


class NoSuitableNodeFoundError(NodeSearchError):
    def __init__(self, node_stack: List[ast.AST]) -> None:
        self.node_stack = node_stack
        super().__init__(
            'no_suitable_node_found',
            "No suitable node found (stack: {})".format(
                " > ".join(type(x).__name__ for x in node_stack),
            ),
        )


class NoSupportedNodeFoundError(NodeSearchError):
    def __init__(self, node_stack: List[ast.AST]) -> None:
        self.node_stack = node_stack
        super().__init__(
            'no_supported_node_found',
            "No supported nodes found (stack: {})".format(
                " > ".join(type(x).__name__ for x in node_stack),
            ),
        )


class NodeFinder(ast.NodeVisitor):
    def __init__(self, position: Position, node_types: Tuple[Type[ast.AST], ...]) -> None:
        self.target_position = position
        self.target_node_types = node_types

        self.node_stack = []  # type: List[ast.AST]

        self.found = False

    def get_filtered_stack(self) -> List[ast.AST]:
        """
        In some cases we want to skip over the actual place in the AST and
        onwards to a parent node. This filtering achieves that.

        For convenience we also return the stack in reversed order, so that the
        nodes closest to the target position are at the front of the list.
        """
        reversed_stack = list(reversed(self.node_stack))

        offset = 0
        for index, node in enumerate(reversed_stack):
            if (
                # Skip upwards past attributes
                isinstance(node, ast.Attribute)
                or (
                    # Skip upwards past calls when our position was on an
                    # attribute to the left of the actual call. For example: in
                    # `func(foo.bar.baz(4))` we want to wrap the call to `baz`
                    # only when on `baz` itself or within its parentheses.
                    # Otherwise we want to wrap `func`. This is somewhat
                    # complicated by the Python AST including `foo.bar.baz` in
                    # the `Call` node, hence this logic.
                    isinstance(node, ast.Call)
                    and node.func in reversed_stack[:index]
                )
            ):
                offset = index + 1
            else:
                break

        return reversed_stack[offset:]

    def _check_not_in_body(self, prev_node: Optional[ast.AST], node: ast.AST) -> None:
        if not prev_node:
            return

        body = getattr(node, 'body', None)
        if not body or not isinstance(body, list):
            return

        if prev_node in body:
            # Don't infer upwards from a function definition or if
            # statement body to the container.
            raise NoSuitableNodeFoundError(self.node_stack)

    @property
    def found_node(self) -> ast.AST:
        if not self.found:
            raise NoNodeFoundError()

        reversed_stack = self.get_filtered_stack()

        for node, prev_node in zip(
            reversed_stack,
            [None, *reversed_stack],
        ):
            if isinstance(node, self.target_node_types):
                self._check_not_in_body(prev_node, node)
                return node

        raise NoSupportedNodeFoundError(self.node_stack)

    def generic_visit(self, node: ast.AST) -> None:
        if self.found:
            return

        if not hasattr(node, 'lineno'):
            super().generic_visit(node)
            return

        start = Position.from_node_start(node)
        end = Position.from_node_end(node)

        if end < self.target_position:
            # we're clear before the target
            return

        if start > self.target_position:
            # we're clear after the target
            return

        # we're on the path to finding the desired node
        self.node_stack.append(node)

        super().generic_visit(node)

        self.found = True


def get_current_indent(asttokens: ASTTokens, node: ast.AST) -> int:
    first_token = _first_token(node)
    lineno = first_token.start[0]

    next_tok = tok = first_token
    while lineno == tok.start[0] and tok.type != token.INDENT:
        next_tok = tok
        tok = asttokens.prev_token(tok)

    return next_tok.start[1]  # type: ignore



import os.path

from setuptools import setup, find_packages  # type: ignore[import]

my_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(my_dir, 'README.md')) as f:
    long_description = f.read()

setup(
    name='tuck',
    version='0.0.10',
    url='https://github.com/PeterJCLaw/tuck',
    project_urls={
        'Issue tracker': 'https://github.com/PeterJCLaw/tuck/issues',
    },
    description="Semi-automated Python formatting.",
    long_description=long_description,
    long_description_content_type='text/markdown',

    packages=find_packages(exclude=['tests']),

    author="Peter Law",
    author_email="PeterJCLaw@gmail.com",

    classifiers=(
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development',
    ),

    install_requires=(
        'asttokens',
    ),
)
